BVI(1)                                                                   BVI(1)



NAME
     bvi - screen-oriented editor for binary files

SYNOPSIS
     bvi [-R] file ...

DESCRIPTION
     bvi is a display-oriented binary editor based on vi(1) texteditor.
     It uses commands similar to the commands of the vi(1), with some
     changes dependent of their different tasks.

     The main differences are:
     
     There are no lines in a binary file. The lines on the screen have no
     relation to the contents of the edited file. So there are no commands
     based on line numbers and no o, O or dd command.

     Also there are no text-specific arrangements like words, paragraphs,
     sentences, sections and so on.

     You cannot insert characters into the file. This makes no sense since
     binary files are mostly executable files or database files and so you
     must not move bytes to other addresses. Also you cannot delete one or
     more character but only overwrite. So there are no i, I, a and x
     command. You can append characters at end of file and you can write a
     new binary file from scratch.

     In the search commands / and ? you can use the \n, \r, \t and \0
     character. Similar to the text search commands there are
     additional hex-search functions \ and #, where you can
     search for any byte value (e.g. \4e 00 d3). In all search functions 
     you can use the regular expression characters . [ ] * but of course
     no ^ and $ (beginning and end of line).

     The screen is divided into a hexadecimal and an ascii section.
     You can toggle between the same byte of the different section
     using the TAB key both in command and input mode.

     At the bottom of the screen there is a statusline, which shows
     the current address of the cursorposition in hexadecimal
	 notation and the value in octal, hexadecimal, decimal and ascii
	 representation.

   Invocation Options
     The following invocation options are interpreted by bvi:

     -R            Readonly mode; the readonly flag is set, preventing
                   accidental overwriting of the file.

     -c  command   Begin editing by executing the specified editor command
                   (usually a search or positioning command).

     The file argument indicates one or more files to be edited.
 
   bvi Modes
     Command        Normal and initial mode.  Other modes return to command
                    mode upon completion.  ESC (escape) is used to cancel a
                    partial command.
 
     Input          Entered by setting any of the following options: A r R.
                    Arbitrary text may then be entered. Input mode
                    is normally terminated with ESC character, or,
                    abnormally, with an interrupt.

     Last line      Reading input for : / ? \ # or !; terminate by typing a
                    carriage return; an interrupt cancels termination.

ENVIRONMENT
     The editor recognizes the environment variable BVIINIT  as  a
     command  (or  list of commands separated by | characters) to
     run when it starts up.  If this variable is  undefined,  the
     editor checks for startup commands in the file ~/.bvirc file,
     which you must own.  However, if there is a .bvirc  owned  by
     you  in  the current directory, the editor takes its startup
     commands from this file - overriding both the file  in  your
     home directory and the environment variable.


COMMAND SUMMARY
     In the descriptions, CR stands for carriage return and ESC stands for the
     escape key.

   Sample commands
     <- v ^ ->         arrow keys move the cursor
     h j k l           same as arrow keys

     u                 undo previous change
     ZZ                exit bvi, saving changes
     :q!CR             quit, discarding changes
     /<text>CR         search for text
     ^U ^D             scroll up or down

   Counts before bvi commands
     Numbers may be typed as a prefix to some commands.  They are interpreted
     in one of these ways.
     screen column        |
     byte of file         G
     scroll amount        ^D  ^U
     repeat effect        most of the rest

   Interrupting, canceling
      ESC       end insert or incomplete cmd
      DEL       (delete or rubout) interrupts

   File manipulation
      ZZ              if file modified, write and exit; otherwise, exit
      :w              write back changes
      :w!             forced write, if permission originally not valid
      :q              quit
      :q!             quit, discard changes
      :e name         edit file name
      :e!             reedit, discard changes
      :e #            edit alternate file
      :e! #           edit alternate file, discard changes
      :w name         write file name
      :w! name        overwrite file name
      :sh             run shell, then return
      :!cmd           run cmd, then return
      :n              edit next file in arglist
      ^G  :f          show current file and line

   Bit-level operations
      :and n          bitwise 'and' operation with value n
      :or n           bitwise 'or' operation with value n
      :xor n          bitwise 'xor' operation with value n
      :neg            two's complement
      :not            logical negation
      :sl i           shift each byte i bits to the left
      :sr i           shift each byte i bits to the right
      :rl i           rotate each byte i bits to the left
      :rr i           rotate each byte i bits to the right

   Command mode addresses
      :w foo          write the whole current workfile to a file named foo
      :5,10w foo      copy byte 5 through 100 into a file named foo
      :.,.+20w foo    copy the current byte and the next 20 bytes to foo
      :^,'aw foo      write all bytes from the beginning through marker 'a'
      :/pat/,$ foo    search pattern pat and and copy through end of file

   Positioning within file
      ^F          forward screen
      ^B          backward screen
      ^D          scroll down half screen
      ^U          scroll up half screen
      nG          go to the specified character
                    (end default), where n is an decimal address
      /pat        next line matching pat
      ?pat        previous line matching pat
      \hex val    next line matching hex values
      #hex val    previous line matching hex values
      n           repeat last / or ? command
      N           reverse last / or ? command

   Adjusting the screen
      ^L            clear and redraw window
      zCR           redraw screen with current line at top of window
      z-CR          redraw screen with current line at bottom of window
      z.CR          redraw screen with current line at center of window
      /pat/z-CR     move pat line to bottom of window
      ^E            scroll window down 1 line
      ^Y            scroll window up 1 line

   Marking and returning
      ``      move cursor to previous context in HEX section
      ''      move cursor to previous context in ASCII section
      mx      mark current position with the ASCII lower-case letter x
      `x      move cursor to mark x in HEX section
      'x      move cursor to mark x in ASCII section

   Line positioning
      H          top line on screen
      L          last line on screen
      M          middle line on screen
      +          next line
      -          previous line
      CR         return, same as +
      UP or j    next line, same column
      DOWN or k  previous line, same column


   Character positioning
      ^           first character in ASCII section
      0           first character in HEX section
      $           end of screen line
      l or ->     forward
      h or <-     backward
      ^H          same as <- (backspace)
      space       same as -> (space bar)
      fx          find next x
      Fx          find previous x
      tx          move to character prior to next x
      Tx          move to character following previous x
      ;           repeat last f F t or T
      ,           repeat inverse of last f F t or T
      n|          move to column n

   Strings (works similar to the strings(1) command)
      w       forward to next string delimited with a nonprinting char
      b       back to previous string delimited with a nonprinting char
      e       end of current word
      W       forward to next string delimited with a \0 or \n
      B       back to previous string delimited with a nonprinting char
      E       end of a blank-delimited word

   Corrections during insert
      ^H        erase last character (backspace)
      erase     your erase character, same as ^H (backspace)
      kill      your kill character, erase this line of input
      \         quotes your erase and kill characters
      ESC       ends insertion, back to command mode
      DEL       interrupt, terminates insert mode

   Append and replace
      A            append at end of file
      rx           replace single char with x
      R<text>ESC   replace characters

   Miscellaneous Operations
      TAB     toggle between ASCII and HEX section

   Yank and Put
      3y      yank 3 characters
      p       replace text with content of yank buffer
      P       put back at end of file
      "xp     put from buffer x
      "xy     yank to buffer x

   Undo, Redo
      u       undo last change
      .       repeat last change

   Setting Options
      With the :set command you can set options in bvi
      Option      Default  Description

      ignorecase  noic     Ignores letter case in searching
      magic       nomagic  Makes . [ * special in searches
      readonly    noro     If set, write fails unless you use ! after command
      offset      of=0     adds an offset to the diplayed addresses
      scroll      sc=1/2 window
                           Number of lines scrolled by ^U and ^D
      showmode    mo       Displays the statusline on bottom of the screen
      wordlength  wl=4     Length of an ASCII-string found by w, W, b or B
      unixstyle   nous     The representation of ascii characters below 32 is
                           displayed in the statusline as shown in ascii(7)
                           if unset rather in DOS-style (^A)

AUTHOR
       Bvi was developed by Gerhard Buergmann, Vienna, Austria
       Email: Gerhard.Buergmann@altavista.net
	   http://bvi.linuxbox.com/

SEE ALSO
       vi(1), strings(1), ascii(5)

