/* COMM.C - Routines to parse and execute "command line" commands, such as
 * search or colon commands.
 *
 * 1996-01-16 created;
 * 1998-03-14 V 1.0.1
 * 1999-02-03 V 1.1.0
 * 1999-04-07 V 1.1.1
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * Copyright 1996-1999 by Gerhard Buergmann
 * Gerhard.Buergmann@altavista.net
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */


#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>

#ifdef __MSDOS__
#	include "doscur.h"
#	include <alloc.h>
#	include <io.h>
#	define WRITE (O_WRONLY|O_CREAT|O_TRUNC|O_BINARY)
#	define APPEND (O_WRONLY|O_APPEND|O_BINARY)
#else
#ifdef NCURSES
#	include <ncurses.h>
#else
#	include <curses.h>
#endif
#	include <unistd.h>
#	define WRITE (O_WRONLY|O_CREAT|O_TRUNC)
#	define APPEND (O_WRONLY|O_APPEND)
#endif

#include "bvi.h"
#include "set.h"

#define CMDLNG(a,b)		(len <= a && len >= b)
#define strsave strdup
#define	CMDSZ	256		/* size of the command buffer */
#define	RBSIZE	1024
#define MAXNAME	10		/* size of a : command name */

#define LSHIFT	1
#define RSHIFT	2
#define LROTATE	3
#define RROTATE	4
#define AND	5
#define OR	6
#define XOR	7
#define NEG	8
#define NOT	9

static	char	*altfile = NULL;	/* alternate file */
static	struct	stat buf;

char	*nowrtmsg  = "No write@since last change (:%s! overrides)";
char	*morefiles = "more files@to edit";
char	*ambigous  = "Ambigous|Too many file names";
char	*ambvalue  = "Ambigous|Too many values";
char	*extra     = "Extra chars|Extra characters at end of command";
char	*noaddr    = "No address allowed@on this command";
char	*noval     = "No value@for binary operation";
int		repl_count = -1;
int		addr_flag;

PTR		start_addr;
PTR		end_addr;

extern	char	*name;			/* actual filename */
extern	char	**files;		/* used for "next" and "rewind" */
extern	int		numfiles, curfile;
extern	int		errno;

static	char	getcbuff[RBSIZE];
static	char	*getcnext = NULL;
static	char	oldbuf[CMDSZ];		/** for :!! command **/

void
stuffin(s)
char	*s;
{
	if (s == NULL) {		/* clear the stuff buffer */
		getcnext = NULL;
		return;
	}
	if (getcnext == NULL) {
		strcpy(getcbuff, s);
		getcnext = getcbuff;
	} else
		strcat(getcbuff, s);
}


int
vgetc()
{
	int	nextc;

	if (getcnext != NULL) {
		nextc = *getcnext++;
		if (*getcnext == '\0') {
			*getcbuff = '\0';
			getcnext = NULL;
		}
		return(nextc);
	}
	return getch();
}



void
clear_marks()
{
	int n;

	for (n = 0; n < 26; markbuf[n++] = NULL);
	undo_count = 0;
	last_motion = mem;
}


/*
 * docmdline() - handle a colon command
 *
 * Handles a colon command received interactively by getcmdln() or from
 * the environment variable "BVIINIT" (or eventually .bvirc).
 */
void
docmdline(cmdline)
	char	*cmdline;
{
	char	buff[CMDSZ];
	char	cmdbuf[CMDSZ];
	char	cmdname[MAXNAME];
	char	*cmd;
	char	*p;
	int		len, n, ok;
	int		force = 0;
	int		c_argc = 0;
	char	*c_argv[9];
	int		saveflag;

	if (cmdline == NULL) return;
	if (*cmdline == '"') return; 	/** comment **/
	if (strlen(cmdline) > CMDSZ - 2) {
		emsg("Command too long");
		return; }

	strcpy(buff, cmdline);

	/* skip any initial white space */
	cmd = buff;

	/* With no address given, we start at the beginning of the file and
	 * go to the end of the file (not line like in vi).
	 */

	addr_flag = 0;
	start_addr = mem;
	end_addr = maxpos - 1;
	SKIP_WHITE
	if (*cmd == '%') {
		cmd++;
	} else {
		if ((start_addr = calc_addr(&cmd, mem)) == NULL) {
			return;
		}
		if (*cmd == ',') {
			cmd++;
			addr_flag++;
			SKIP_WHITE
			if ((end_addr = calc_addr(&cmd, maxpos - 1)) == NULL) {
				return;
			}
		} else {
			if (addr_flag) {
				/* if we have only one address */
				end_addr = start_addr;
			}
		}
	}
	SKIP_WHITE

	if (start_addr > (end_addr + 1)) {
		emsg("Addr1 > addr2|First address exceeds second");
		return;
	}
	if ((end_addr + 1) > maxpos) {
		emsg("Not that many bytes@in buffer");
		return;
	}
	if (start_addr < mem) {
		sprintf(string, "Negative address@- first byte is %d", P(P_OF));
		emsg(string);
		return;
	}

	SKIP_WHITE

/**** End of address range calculation ****/

	if (*cmd == '\0') {
		setpage(end_addr);
		return;
	}
	strcpy(cmdbuf, cmd);	/* save the unmodified command */

	if (*cmd == '!') {
		if (*(cmdbuf + 1) == '\0') {
			emsg("Incomplete shell escape command@- use 'shell' to get a shell");
			return;
			}
		if (*(cmdbuf + 1) == '!') {
			if (oldbuf[0] != '\0') {
				strcpy(cmdbuf + 1, oldbuf);
				msg(oldbuf);
			} else {
				emsg("No previous command@to substitute for !");
				return;
			}
		} else	sprintf(oldbuf, "\n%s\n", cmdbuf + 1);

		if (P(P_AW)) {
			save(name, mem, maxpos, WRITE);
			edits = 0;
		}
		if (edits) msg("[No write]|[No write since last change]");
		savetty();
		endwin();
		system(cmdbuf + 1);
		fprintf(stderr, "[Hit return to continue]");
		getchar();
		doupdate();
		repaint();
		clearstr();
		return;
	}

	n = 0;
	while (*cmd >= 'a' && *cmd <= 'z') {
		cmdname[n++] = *cmd++;
		if (n == MAXNAME)  {
			sprintf(string, "What?|%s: Not an editor command", cmdbuf);
			emsg(string);
			return;
			}
	}
	cmdname[n] = '\0';
	if (*cmd == '!') {
		force = 1;
		cmd++;
		}
	len = strlen(cmdname);
	SKIP_WHITE

	if (!strncmp("substitute", cmdname, len) && CMDLNG(10,1)) {
		repl_count = do_substitution(*cmd, cmd + 1);
		if (repl_count == -1) {
			emsg("No previous substitute re|No previous substitute regular expression");
			return;	/* No prev subst */
			}
		repaint();
		if (!repl_count) {
			emsg("Fail|Substitute pattern matching failed");
		} else if (repl_count > 1) {
			sprintf(string, "%d subs|%d substitutions", repl_count, repl_count);
			msg(string);
		}
		return;
	}

	if ((!strncmp("write", cmdname, len) && CMDLNG(5,1)) || !strcmp("wq", cmdname)) {
		if (*cmd == '>') {
			if (*(++cmd) == '>') {
				cmd++;
				saveflag = APPEND;
			} else {
				emsg("Write forms are 'w' and 'w>>'");
				return;
			}
		} else if (*cmd == '!') {
			emsg("Not yet implemented");
			return;
		} else {
			saveflag = WRITE;
		}
		SKIP_WHITE

		c_argc = 0;
		if ((c_argv[c_argc] = strtok(cmd, " \t")) != NULL) c_argc++;
		while ((c_argv[c_argc] = strtok(NULL, " \t")) != NULL) c_argc++;

		if (c_argc > 1) { 
			emsg(ambigous);
			return; }
		if (c_argc == 1) {
		    /* change '%' to Filename */
		    while ((p = strchr(c_argv[0], '%')) != NULL && *(p-1) != '\\') {
		        if (name == NULL) {
		            emsg("No filename@to substitute for %");
		            return;
		        }
		        *p = '\0';
		        strcpy (oldbuf, c_argv[0]);
		        strcat (oldbuf, name);
		        strcat (oldbuf, p + 1);
				c_argv[0] = oldbuf;
		    }
			/* change '#' to Altname */
			while ((p = strchr(cmd, '#')) != NULL && *(p - 1) != '\\') {
				if (altfile == NULL) {
					emsg("No alternate filename@to substitute for #");
					return;
				}
				*p = '\0';
				strcpy (cmdbuf, cmd);
				strcat (cmdbuf, altfile);
				strcat (cmdbuf, p + 1);
				strcpy(cmd, cmdbuf);
				msg(cmd);			/*repeat */
			}
		}
		if (name == NULL && c_argc != 0) name = strsave(c_argv[0]);
		if (force) {
			if (c_argc == 0) ok = save(name, start_addr, end_addr, saveflag);
				else	ok = save(c_argv[0], start_addr, end_addr, saveflag);
		} else {
			if (c_argc == 0) {
				if (P(P_RO)) {
					sprintf(string, "\"%s\" File is read only", name);
					emsg(string);
					return;
				} else ok = save(name, start_addr, end_addr, saveflag);
			} else {
				if (!stat(c_argv[0], &buf)) {
					if (saveflag == WRITE) {
						sprintf(string, "File exists@- use \"%s! %s\" to overwrite", cmdname, c_argv[0]);
						emsg(string);
						return;
					} else {
						ok = save(c_argv[0], start_addr, end_addr, saveflag);
					}
				} else {
					if (saveflag == APPEND) {
						emsg("No such file");
						return;
					} else {
						ok = save(c_argv[0], start_addr, end_addr, saveflag);
					}
				}
			}
		}
		if (!strcmp("wq", cmdname)) {
			if (ok) quit();
		}
		return;
	}

	c_argc = 0;
	cmd = strtok(cmdbuf, " \t");
	while ((c_argv[c_argc] = strtok(NULL, " \t")) != NULL) c_argc++;

	if (!strncmp("set", cmdname, len) && CMDLNG(3, 2)) {
		if (addr_flag) { emsg(noaddr); return; }
		if (c_argc == 0) doset(NULL);
		else
			for (n = 0; n < c_argc; n++) {
				if (doset(c_argv[n])) return;
				}
	} else if (!strncmp("args", cmdname, len) && CMDLNG(4, 2)) {
		if (addr_flag) { emsg(noaddr); return; }
		if (c_argc == 0) {
			strcpy(string, "");
			for (n = 0; n < numfiles; n++) {
				if (n == curfile) strcat(string, "[");
				strcat(string, files[n]);
				if (n == curfile) strcat(string, "]");
				strcat(string, " ");
			}
			msg(string);
		} else {
			/* this is the place to set new arguments **********/
			/*
			name = strdup(c_argv[0]);
			files = &(c_argv[0]);
			numfiles = c_argc;
			*/
		}
	} else if (!strncmp("source", cmdname, len) && CMDLNG(6, 2)) {
		if (addr_flag) { emsg(noaddr); return; }
    	if (c_argc > 1) { emsg(ambigous); return; }
		if (c_argc == 0) { emsg("Missing filename"); return; }
		if (read_rc(c_argv[0])) sysemsg(c_argv[0]);
		refresh();
	} else if (!strncmp("cd", cmdname, len) && CMDLNG(2, 2)) {
		if (addr_flag) { emsg(noaddr); return; }
    	if (c_argc > 1) { emsg(ambigous); return; }
		if (!force) {
			if (edits) {
				if (P(P_AW)) {
					save(name, mem, maxpos, WRITE);
					edits = 0;
				} else {
					sprintf(string, nowrtmsg, "cd");
					emsg(string);
					return;
				}
			}
		}
		if (c_argc == 0) {
			sprintf(string, "%c", DELIM);
			c_argv[0] = string;
			}
		if (chdir(c_argv[0])) sysemsg(c_argv[0]);
	} else if (!strncmp("edit", cmdname, len) && CMDLNG(4, 1)) {
		if (addr_flag) { emsg(noaddr); return; }
		/*
		if (start_addr != mem || end_addr != (maxpos - 1)) {
			emsg("No address allowed@on this command");
			return; 
			}
		*/
		if (c_argc > 1) { emsg(ambigous); return; }
		doecmd(c_argv[0], force);
	/*
	 * The command ":e#" gets expanded to something like ":efile", so
	 * detect that case here.
	 */
	} else if (*cmd == 'e' && c_argc == 0) {
		if (cmd[1] == '!')
			(void) doecmd(&cmd[2], TRUE);
		else
			(void) doecmd(&cmd[1], FALSE);
	} else if (!strncmp("file", cmdname, len) && CMDLNG(4, 1)) {
		if (addr_flag) { emsg(noaddr); return; }
		if (c_argc != 0) name = strsave(c_argv[0]);
		fileinfo(name);
	} else if (!strncmp("read", cmdname, len) && CMDLNG(4, 1)) {
		if (addr_flag) { emsg(noaddr); return; }
		/* read reads a file at EOF */
		if (c_argc > 1) { emsg(ambigous); return; }
		if (c_argc == 0) c_argv[0] = name;
		if (!addfile(c_argv[0])) {
			edits = TRUNC;
			undosize = filesize;
			repaint();
		}
	} else if (!strncmp("xit", cmdname, len) && CMDLNG(3, 1)) {
		if (addr_flag) { emsg(noaddr); return; }
		if (c_argc != 0) { emsg(extra); return; }
		do_exit();
	} else if (!strncmp("next", cmdname, len) && CMDLNG(4, 1)) {
		if (addr_flag) { emsg(noaddr); return; }
		if (c_argc != 0) { emsg(extra); return; }
		if (!force) {
			if (edits) {
				if (P(P_AW)) {
					save(name, mem, maxpos, WRITE);
					edits = 0;
				} else {
					sprintf(string, nowrtmsg, "next");
					emsg(string);
					return;
				}
			}
		}
		if ((curfile + 1) < numfiles) {
			if (force)  stuffin(":e! ");
			else		stuffin(":e ");
			stuffin(files[++curfile]);
			stuffin("\n");
		} else
			emsg("No more files@to edit!");
	} else if (!strncmp("rewind", cmdname, len) && CMDLNG(6, 3)) {
		if (addr_flag) { emsg(noaddr); return; }
		if (c_argc != 0) { emsg(extra); return; }
		if (numfiles <= 1)		/* nothing to rewind */
			return;
		if (!force) {
			if (edits) {
				if (P(P_AW)) {
					save(name, mem, maxpos, WRITE);
					edits = 0;
				} else {
					sprintf(string, nowrtmsg, "rewind");
					emsg(string);
					return;
				}
			}
		}
		curfile = 0;
		if (force)	stuffin(":e! ");
			else	stuffin(":e ");
		stuffin(files[0]);
		stuffin("\n");
	} else if (!strncmp("version", cmdname, len) && CMDLNG(7, 2)) {
		extern	char	*version;
		if (addr_flag) { emsg(noaddr); return; }
        if (c_argc != 0) { emsg(extra); return; }
		msg(version);
	} else if (!strncmp("shell", cmdname, len) && CMDLNG(5, 2)) {
		if (addr_flag) { emsg(noaddr); return; }
        if (c_argc != 0) { emsg(extra); return; }
		do_shell();
		clear();
		repaint();
		clearstr();
	} else if (!strncmp("quit", cmdname, len) && CMDLNG(4, 1)) {
		if (addr_flag) { emsg(noaddr); return; }
		if (c_argc != 0) { emsg(extra); return; }
		if (!force) {
			if (edits) {
				sprintf(string, nowrtmsg, "quit");
				emsg(string);
			} else {
				if ((curfile + 1) < numfiles) {
					sprintf(string, "%d %s", numfiles - curfile - 1, morefiles);
					emsg(string);
				} else
					quit();
			}
		} else quit();
	} else if (!strncmp("sleft", cmdname, len) && CMDLNG(5, 2)) {
		if (c_argc == 1) {
			do_logic(LSHIFT, c_argv[0]);
		} else if (c_argc == 0) {
			do_logic(LSHIFT, "1");
    	} else { emsg(ambvalue); return; }
	} else if (!strncmp("sright", cmdname, len) && CMDLNG(6, 2)) {
		if (c_argc == 1) {
			do_logic(RSHIFT, c_argv[0]);
		} else if (c_argc == 0) {
			do_logic(RSHIFT, "1");
    	} else { emsg(ambvalue); return; }
	} else if (!strncmp("rleft", cmdname, len) && CMDLNG(5, 2)) {
		if (c_argc == 1) {
			do_logic(LROTATE, c_argv[0]);
		} else if (c_argc == 0) {
			do_logic(LROTATE, "1");
    	} else { emsg(ambvalue); return; }
	} else if (!strncmp("rright", cmdname, len) && CMDLNG(6, 2)) {
		if (c_argc == 1) {
			do_logic(RROTATE, c_argv[0]);
		} else if (c_argc == 0) {
			do_logic(RROTATE, "1");
    	} else { emsg(ambvalue); return; }
	} else if (!strcmp("and", cmdname)) {
		if (c_argc == 1) {
			do_logic(AND, c_argv[0]);
		} else if (c_argc == 0) {
			emsg(noval); return;
    	} else { emsg(ambvalue); return; }
	} else if (!strcmp("or", cmdname)) {
		if (c_argc == 1) {
			do_logic(OR, c_argv[0]);
		} else if (c_argc == 0) {
			emsg(noval); return;
    	} else { emsg(ambvalue); return; }
	} else if (!strcmp("xor", cmdname)) {
		if (c_argc == 1) {
			do_logic(XOR, c_argv[0]);
		} else if (c_argc == 0) {
			emsg(noval); return;
    	} else { emsg(ambvalue); return; }
	} else if (!strcmp("neg", cmdname)) {
		if (c_argc != 0) { emsg(extra); return; }
		do_logic(NEG, "255");
	} else if (!strcmp("not", cmdname)) {
		if (c_argc != 0) { emsg(extra); return; }
		do_logic(NOT, "255");
	} else {
		sprintf(string, "What?|%s: Not an editor command", cmd);
		emsg(string);
	}
}


int
do_logic(mode, str)
	int		mode;
	char	*str;
{
	int		a, b;
	int		value;
	int		n;
	char	*err_str ="Invalid value@for bit manipulation";

	if (mode==LSHIFT || mode==RSHIFT || mode==LROTATE || mode==RROTATE) {
		value = atoi(str);
		if (value < 1 || value > 8) {
			emsg(err_str);
			return 1;
		}
	} else {
		if (strlen(str) == 8) {
			value = strtol(str, NULL, 2);
			for (n = 0; n < 8; n++) {
				if (str[n] != '0' && str[n] != '1') {
					value = -1;
					break;
				}
			}
		} else if (str[0] == 'b' || str[0] == 'B') {
			value = strtol(str + 1, NULL, 2);
		} else if (str[0] == '0') {
			value = strtol(str, NULL, 16);
			for (n = 0; n < strlen(str); n++) {
				if (!isxdigit(str[n])) {
					value = -1;
					break;
				}
			}
		} else {
			value = atoi(str);
		}
		if (value < 0 || value > 255) {
			emsg(err_str);
			return 1;
		}
	}
	if (undo_count = alloc_buf(end_addr - start_addr + 1, &undo_buf)) {
		memcpy(undo_buf, start_addr, undo_count);
	}
	undo_start = start_addr;
	edits = EDIT;
	while(start_addr <= end_addr) {
		a = *start_addr;
		a &= 0xff;
		switch(mode) {
		case LSHIFT:	a <<= value;
						break;
		case RSHIFT:	a >>= value;
						break;
		case LROTATE:	a <<= value;
						b = a >> 8;
						a |= b;
						break;
		case RROTATE:	
						b = a << 8;
						a |= b;
						a >>= value; 
						/*
						b = a << (8 - value);
						a >>= value; 
						a |= b;
						*/
						break;
		case AND:		a &= value;
						break;
		case OR:		a |= value;
						break;
		case XOR:		
		case NOT:		a ^= value;
						break;
		case NEG:		a ^= value;
						a++;		/* Is this true */
						break;
		}
		*start_addr++ = (char)(a & 0xff);
	}
	repaint();
	return(0);
}


int
addfile(fname)
	char	*fname;
{
	int		fd;
	size_t	oldsize;

	if (stat(fname, &buf)) {
		sysemsg(fname);
		return 1;	}
	if ((fd = open(fname, O_RDONLY)) == -1) {
		sysemsg(fname);
		return 1;	}
	oldsize = filesize;
	if (enlarge(buf.st_size)) return 1;
	if (read(fd, mem + filesize, buf.st_size) == -1) {
		sysemsg(fname);
		return 1;	}
	filesize += buf.st_size;
	maxpos = mem + filesize;
	close(fd);
	setpage(mem + oldsize);

	return 0;
}


void
do_exit()
{
	if (edits) {
			if (!save(name, mem, maxpos, WRITE)) return;
		}
	if ((curfile + 1) < numfiles) {
		sprintf(string, "%d %s", numfiles - curfile - 1, morefiles);
		emsg(string);
	} else
		quit();
}


int
doecmd(arg, force)
	char	*arg;
	int		force;
{
	if (!force && edits) {
		sprintf(string, nowrtmsg, "edit");
		emsg(string);
		if (altfile)
			free(altfile);
		altfile = strsave(arg);
		return FALSE;
	}
	if (arg != NULL) {
		if (name != NULL && strcmp(arg, name) == 0) {
			if (!edits || (edits && !force))
				return TRUE;
		}
		if (altfile) {
			free(altfile);
		}
		altfile = name;
		name = strsave(arg);
	}
	if (name == NULL) {
		emsg("No file|No current filename");
		return FALSE;
	}

	edits = 0;
	filesize = load(name);
	return TRUE;
}


int
getcmdstr(p, x)
	char    *p;
	int		x;
{
	int		c;
	int		n;
	char	*buff, *q;

	buff = p;
	move(maxy, x);
    do {
        switch (c = vgetc()) {
		case CTRL('H'):
		case KEY_BACKSPACE:
		case KEY_LEFT:
            if (p > buff) {
                p--;
				move(maxy, x);
				n = x;
                for (q = buff; q < p ;q++)  {
                    addch(*q);
					n++;
				}
				addch(' ');
				move(maxy, n);
            } else {
				*buff = '\0';
                msg("");
                return 1;
            }
            break;
        case ESC:           /* abandon command */
			*buff = '\0';
            msg("");
            return  1;
        case NL:            /* done reading the line */
        case CR:
            break;
        default:        /* a normal character */
            addch(c);
            *p++ = c;
            break;
        }
		refresh();
    } while (c != NL && c != CR);

    *p = '\0';
	/*
	move(maxy, x);
	sprintf(oldbuf, "%c%s", ch, buff);
	msg(oldbuf);
	*/
	return 0;
}


void
clearstr()
{
	int n;

	move(maxy, 0);
	for (n = 0; n < maxx; n++) addch(' ');
	move(maxy, 0);
}


/**** displays an error message *****/
void
emsg(s)
	char	*s;
{
	int	cnt;

	if (P(P_EB)) beep();
	clearstr();
	attrset(A_REVERSE);
	cnt = outmsg(s);
	attrset(A_NORMAL);
	if (cnt >= (maxx - 25)) {		/* 25 = status */
		addch('\n');
		wait_return(); }
}


/*** System error message *****/
void
sysemsg(s)
	char	*s;
{
	char	string[256];

	sprintf(string, "%s: %s", s, sys_errlist[errno]);
	emsg(string);
}


/*** displays mode if showmode set *****/
void
smsg(s)
	char	*s;
{
	if (P(P_MO)) {
		msg(s);
		setcur();
	}
}


/************* displays s on status line *****************/
void
msg(s)
	char	*s;
{
	clearstr();
	if (outmsg(s) >= (maxx - 25)) {		/* 25 = status */
		addch('\n');
		wait_return(); }
}


outmsg(s)
	char	*s;
{
	char	*poi;
	int		cnt = 0;

	move(maxy, 0);
	poi = strchr(s, '|');

	if (P(P_TE)) {
		poi = s;
		while (*poi != '\0' && *poi != '@' && *poi != '|') {
			addch(*poi++);
			cnt++; }
	} else {
    	if (poi) poi++;
			else poi = s;
		while (*poi) {
			if (*poi == '@') addch(' ');
				else addch(*poi);
			poi++;
			cnt++;
		}
	}
	return cnt;
}


void
wait_return()
{
	register char	c;

	clearstr();
	attrset(A_REVERSE);
	mvaddstr(maxy, 0, "[Hit return to continue]");
	attrset(A_NORMAL);
	refresh();
	do {
		c = getch();
	} while (c != CR && c != NL && c != ' ' && c != ':');
	clear();
	repaint();
	clearstr();
}
