/* IO.C - file in/out and alloc subroutines for BVI
 *
 * 1996-02-28 V 1.00
 * 1998-04-12 V 1.01
 * 1999-01-20 V 1.10
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * Copyright 1996-1999 by Gerhard Buergmann
 * Gerhard.Buergmann@altavista.net
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <curses.h>
#include "bvi.h"
#include "set.h"

int		filemode;
static	struct	stat	buf;
char	*terminal;


/*********** Save the patched file ********************/
int
save(fname, start, end, flags)
	char	*fname;
	char	*start;
	char	*end;
	int		flags;
{
	int		fd;
	char	string[255];
	char	*newstr;
	long	filesize;

	if (!fname) {
		emsg("No file|No current filename");
		return 0;
		}
	if (stat(fname, &buf) == -1) {
		newstr = "[New file] ";
	} else {
		if (S_ISDIR(buf.st_mode)) {
			sprintf(string, "\"%s\" Is a directory", fname);
			msg(string);
			return 0;
		} else if (S_ISCHR(buf.st_mode)) {
			sprintf(string, "\"%s\" Character special file", fname);
			msg(string);
			return 0;
		} else if (S_ISBLK(buf.st_mode)) {
			sprintf(string, "\"%s\" Block special file", fname);
			msg(string);
			return 0;
		}
		newstr = "";
	}

	if ((fd = open(fname, flags, 0666)) < 0) {
        sysemsg(fname);
		return 0;
	}
	filesize = end - start + 1L;

	if (write(fd, start, filesize) != filesize) {
		sysemsg(fname);
		close(fd);
		return 0;
	}
	close(fd);
	edits = 0;
	sprintf(string, "\"%s\" %s%d bytes", fname, newstr, filesize);
	msg(string);
	return 1;
}


/* loads a file, returns the filesize */
long
load(fname)
	char	*fname;
{
	int		fd = -1;
	char	string[MAXCMD];

	buf.st_size = 0L;
	if (fname != NULL) {
		if (stat(fname, &buf) == -1) {
			filemode = NEW;
		} else if (S_ISDIR(buf.st_mode)) {
			filemode = DIRECTORY;
		} else if (S_ISCHR(buf.st_mode)) {
			filemode = CHARACTER_SPECIAL;
		} else if (S_ISBLK(buf.st_mode)) {
			filemode = BLOCK_SPECIAL;
		} else if (S_ISREG(buf.st_mode)) {
			if ((fd = open (fname, O_RDONLY)) > 0) {
				filemode = REGULAR;
				if (access(fname, W_OK)) {
					P(P_RO) = TRUE;
					params[P_RO].flags |= P_CHANGED;
					}
			} else {
	            sysemsg(fname);
				filemode = ERROR;
			}
		}
	} else {
		filemode = NEW;
	}
	if (mem != NULL) free(mem);
	if (filemode == REGULAR) {
		memsize = buf.st_size + 1000L;
	} else {
		memsize = 1000L;
	}
	if ((mem = (char *)malloc(memsize)) == NULL) {
		move(maxy, 0);
		endwin();
		printf("Out of memory\n");
		exit(0);
		}
	clear_marks();
	if (filemode == REGULAR) {
		if (read(fd, mem, buf.st_size) != buf.st_size) {
            sysemsg(fname);
			filemode = ERROR;
		}
	}
	if (fd > 0) close(fd);
	if (filemode == REGULAR) {
		filesize = buf.st_size;
	} else {
		filesize = 0L;
	}
	if (fname != NULL) {
		switch (filemode) {
		case NEW:
			sprintf(string, "\"%s\" [New File]", fname);
			break;
		case REGULAR:
			sprintf(string, "\"%s\" %s%d bytes", fname,
				P(P_RO) ? "[Read only] " : "",
				filesize);
			break;
		case DIRECTORY:
			sprintf(string, "\"%s\" Directory", fname);
			break;
		case CHARACTER_SPECIAL:
			sprintf(string, "\"%s\" Character special file", fname);
			break;
		case BLOCK_SPECIAL:
			sprintf(string, "\"%s\" Block special file", fname);
			break;
		}
		if (filemode != ERROR) msg(string);
	}
	pagepos = mem;
	maxpos = mem + filesize;
	loc = HEX;
	x = AnzAdd; y = 0;
	repaint();
	return(filesize);
}


void
bvi_init()
{
	char    *initstr;
	char    rcpath[255];

	terminal = getenv("TERM");
	shell = getenv("SHELL");
	if (shell == NULL || *shell == '\0')
		shell = "/bin/sh";

	if ((initstr = getenv("BVIINIT")) != NULL) {
			docmdline(initstr);
			return; }

	strcpy(rcpath, getenv("HOME"));
	strcat(rcpath, "/.bvirc");
	if (stat(rcpath, &buf) == 0) {
		if (buf.st_uid == getuid()) read_rc(rcpath);
		}

	strcpy(rcpath, ".bvirc");
	if (stat(rcpath, &buf) == 0) {
		if (buf.st_uid == getuid())	read_rc(rcpath);
		}
}


int
enlarge(add)
	long	add;
{
	char	*newmem;
	long	savecur, savepag, savemax;

	savecur = curpos - mem;
	savepag = pagepos - mem;
	savemax = maxpos - mem;

	newmem = realloc(mem, memsize + add);
	if (newmem == NULL) {
		emsg("Out of memory");
		return 1; }

	mem = newmem;
	memsize += add;
	curpos = mem + savecur;
	pagepos = mem + savepag;
	maxpos = mem + savemax;
	current = curpos + 1;

	return 0;
}


void
do_shell()
{
	addch('\n');
	savetty();
	system(shell);
	resetty();
}


char *
strdup(s)
	char	*s;
{
	char    *p;
	unsigned n;

	n = strlen(s) + 1;
	if ((p = (char *)malloc(n)) != NULL)
		memcpy(p, s, n);
	return (p);
}


long
alloc_buf(n, buffer)
	long	n;
	char	**buffer;
{
	if (*buffer != NULL) free(*buffer);
    if ((*buffer = (char *)malloc(n)) == NULL) {
		emsg("No buffer space available");
		return 0L;
	} else return n;
}
